from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import math


class EtchAgent(LemonadeAgent):
    def setup(self):
        pass

    def get_action(self):
        try:
            last_opp_1 = self.get_opp1_action_history()[-1]
            last_opp_2 = self.get_opp2_action_history()[-1]
            action1 = (last_opp_1 + 6) % 12
            vals = 0
            for i in range(self.get_opp1_action_history()):
                val = math.abs(self.get_opp1_action_history()[
                               i] - self.get_opp2_action_history()[i])
                if (val > 6):
                    distance_away = val / 2
                else:
                    distance_away = (12 - val) / 2

                if (self.get_opp1_action_history()[i] > self.get_opp2_action_history()[i]):
                    bigger_val = self.get_opp1_action_history()[i]
                else:
                    bigger_val = self.get_opp2_action_history()[i]

                vals = vals + (bigger_val + distance_away) % 12

            val1 = math.abs(self.get_opp1_action_history()
                            [-1] - self.get_opp2_action_history()[-1])
            if (val1 > 6):
                distance_away1 = val / 2
            else:
                distance_away1 = (12 - val) / 2

            if (self.get_opp1_action_history()[-1] > self.get_opp2_action_history[-1]):
                bigger_val1 = self.get_opp1_action_history()[i]
            else:
                bigger_val1 = self.get_opp2_action_history()[i]

            val1 = val1 + (bigger_val1 + distance_away1) % 12

            action = (vals / len(self.get_opp1_action_history()) + val1) / 2

        except:
            action = random.choice([2])
        return action

    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = EtchAgent("Etch")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = EtchAgent(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                EtchAgent("Agent_1"),
                EtchAgent("Agent_2"),
                EtchAgent("Agent_3"),
                EtchAgent("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.lsvm_agent import MyLSVMAgent
from agt_server.local_games.lsvm_arena import LSVMArena
import time

class JumpBidder(MyLSVMAgent):
    def setup(self):
        pass

    def get_bids(self):
        min_bids = self.get_min_bids()
        valuations = self.get_valuations() 
        bids = {}
        for good in valuations: 
            if valuations[good] > min_bids[good]:
                bids[good] = min_bids[good] + (0.5 * (valuations[good] - min_bids[good]))
        return bids
    
    def update(self):
        pass


################### SUBMISSION #####################
my_agent_submission = JumpBidder("JumpBidder")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    agent = JumpBidder("Truthful")
    arena = LSVMArena(
        num_cycles_per_player = 5,
        timeout=1,
        # local_save_path="saved_games",
        players=[
            agent,
            JumpBidder("Agent_1"),
            JumpBidder("Agent_2"),
            JumpBidder("Agent_3"),
            JumpBidder("Agent_4"), 
            JumpBidder("Agent_5"), 
            JumpBidder("Agent_6"), 
        ]
    )
    
    start = time.time()
    arena.run()
    end = time.time()
    print(f"{end - start} Seconds Elapsed")



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np


class Sticky(LemonadeAgent):

    def setup(self):
        self.move = random.randint(0, 11)  # choose a random initial location

    def get_action(self):
        # if this is not the first turn, check if opponents are trying to maximize or minimize distance
        if len(self.get_action_history()) > 0:
            opp_1_dist = (self.get_last_action() - self.get_opp1_action_history()[-1]) % 12  # calculate distance from opponent 1
            opp_2_dist = (self.get_last_action() - self.get_opp2_action_history()[-1]) % 12  # calculate distance from opponent 2

            # if both opponents are close (distance <= 3), assume they are trying to minimize distance
            if opp_1_dist <= 3 and opp_2_dist <= 3:
                # move to the opposite location (over the diagonal)
                self.move = (self.get_last_action() + 6) % 12

            # otherwise, assume they are trying to maximize distance and maintain current location

        # append the chosen action to the history and return it
        return self.move

    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = Sticky("Elmers")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = Sticky(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                Sticky("Agent_1"),
                Sticky("Agent_2"),
                Sticky("Agent_3"),
                Sticky("Agent_4")
            ]
        )
        arena.run()



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np


class Spinner(LemonadeAgent):
    def get_best_pos(self, a, b):
        ret = (a + b) // 2
        if min(abs((a - ret + 12) % 12), abs((ret - a + 12) % 12)) > 2:
            return ret
        else:
            return (ret + 6) % 12

    def setup(self):
        pass

    def get_action(self):
        try:
            last_opp_1 = self.get_opp1_action_history()[-1]
            last_opp_2 = self.get_opp2_action_history()[-1]
            action = random.choice([random.choice(
                [last_opp_1, last_opp_2]), self.get_best_pos(last_opp_1, last_opp_2)])
        except:
            action = random.choice([0, 4, 8])
        return action

    def update(self):
        pass

################### SUBMISSION #####################
agent_submission = Spinner("Spinner")
####################################################

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = Spinner(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                Spinner("Agent_1"),
                Spinner("Agent_2"),
                Spinner("Agent_3"),
                Spinner("Agent_4")
            ]
        )
        arena.run()



from agt_server.local_games.base import LocalArena
from itertools import combinations
import numpy as np
from collections import defaultdict
import json
from datetime import datetime
import pkg_resources
import random

class LemonadeArena(LocalArena):
    def __init__(self, num_rounds=1000, players=[], timeout=1, handin=False, logging_path = None, save_path = None):
        super().__init__(num_rounds, players, timeout, handin, logging_path, save_path)
        self.game_name = "Lemonade Stand"
        self.valid_actions = list(range(0, 12))

        if not self.handin_mode:
            assert len(self.players) >= 3, "Arena must have at least 3 players"

        for idx in range(len(self.players)):
            if self.handin_mode:
                try:
                    player = self.players[idx]
                    self.game_reports[player.name] = {
                        "action_history": [],
                        "util_history": [],
                        "index": idx,
                        "timeout_count": 0,
                        "global_timeout_count": 0,
                        "disconnected": False
                    }
                except:
                    continue
            else:
                player = self.players[idx]
                self.game_reports[player.name] = {
                    "action_history": [],
                    "util_history": [],
                    "index": idx,
                    "global_timeout_count": 0
                }

        self.results = []
        self.game_num = 1

    def reset_game_reports(self):
        for player in self.players:
            if self.handin_mode:
                try:
                    self.game_reports[player.name]["action_history"] = []
                    self.game_reports[player.name]["util_history"] = []
                except:
                    continue
            else:
                self.game_reports[player.name]["action_history"] = []
                self.game_reports[player.name]["util_history"] = []

    def run(self):
        for p1, p2, p3 in combinations(self.players, 3):
            if self.handin_mode:
                if p1 is None or p2 is None or p3 is None or self.game_reports[p1.name]['disconnected'] or self.game_reports[p2.name]['disconnected'] or self.game_reports[p3.name]['disconnected']:
                    continue
                else:
                    try:
                        self.run_func_w_time(
                            p1.restart, self.timeout, p1.name)
                    except:
                        self.game_reports[p1.name]['disconnected'] = True
                        continue
                    
                    try:
                        self.run_func_w_time(
                            p2.restart, self.timeout, p2.name)
                    except:
                        self.game_reports[p2.name]['disconnected'] = True
                        continue
                    
                    try:
                        self.run_func_w_time(
                            p3.restart, self.timeout, p3.name)
                    except:
                        self.game_reports[p3.name]['disconnected'] = True
                        continue
                    
                    try:
                        self.run_game(p1, p2, p3)
                    except:
                        self.reset_game_reports()
                        continue
            else:
                self.run_func_w_time(p1.restart, self.timeout, p1.name)
                self.run_func_w_time(p2.restart, self.timeout, p2.name)
                self.run_func_w_time(p3.restart, self.timeout, p3.name)
                self.run_game(p1, p2, p3)
        results = self.summarize_results()
        return results

    def calculate_utils(self, p1_action, p2_action, p3_action):
        if p1_action not in self.valid_actions and p2_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [0, 0, 0]
        elif p1_action not in self.valid_actions and p2_action not in self.valid_actions:
            utils = [0, 0, 24]
        elif p1_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [0, 24, 0]
        elif p2_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [24, 0, 0]
        elif p1_action not in self.valid_actions:
            utils = [0, 12, 12]
        elif p2_action not in self.valid_actions:
            utils = [12, 0, 12]
        elif p3_action not in self.valid_actions:
            utils = [12, 12, 0]
        elif p1_action == p2_action and p2_action == p3_action:
            utils = [8, 8, 8]
        elif p1_action == p2_action:
            utils = [6, 6, 12]
        elif p1_action == p3_action:
            utils = [6, 12, 6]
        elif p2_action == p3_action:
            utils = [12, 6, 6]
        else:
            utils = [0, 0, 0]
            actions = [p1_action, p2_action, p3_action]
            sorted_actions = sorted(actions)
            index_map = {action: index for index,
                         action in enumerate(actions)}
            sorted_indices = [index_map[action]
                              for action in sorted_actions]
            u1 = sorted_actions[1] - sorted_actions[0]
            u2 = sorted_actions[2] - sorted_actions[1]
            u3 = 12 + sorted_actions[0] - sorted_actions[2]
            utils[sorted_indices[0]] = u1 + u3
            utils[sorted_indices[1]] = u1 + u2
            utils[sorted_indices[2]] = u2 + u3
        return utils

    def run_helper(self, p1, p2, p3):
        for _ in range(self.num_rounds):
            if self.handin_mode:
                if self.game_reports[p1.name]['disconnected'] or self.game_reports[p2.name]['disconnected'] or self.game_reports[p3.name]['disconnected']:
                    break
                if self.game_reports[p1.name]['timeout_count'] < self.timeout_tolerance:
                    try:
                        self.run_func_w_time(
                            p1.restart, self.timeout, p1.name)
                        p1_action = self.run_func_w_time(
                            p1.get_action, self.timeout, p1.name, -1)
                    except:
                        self.game_reports[p1.name]['disconnected'] = True
                        p1_action = -1
                else:
                    self.game_reports[p1.name]['disconnected'] = True
                    p1_action = -1

                if self.game_reports[p2.name]['timeout_count'] < self.timeout_tolerance:
                    try:
                        self.run_func_w_time(
                            p2.restart, self.timeout, p1.name)
                        p2_action = self.run_func_w_time(
                            p2.get_action, self.timeout, p2.name, -1)
                    except:
                        self.game_reports[p2.name]['disconnected'] = True
                        p2_action = -1
                else:
                    self.game_reports[p2.name]['disconnected'] = True
                    p2_action = -1

                if self.game_reports[p3.name]['timeout_count'] < self.timeout_tolerance:
                    try:
                        self.run_func_w_time(
                            p3.restart, self.timeout, p1.name)
                        p3_action = self.run_func_w_time(
                            p3.get_action, self.timeout, p3.name, -1)
                    except:
                        self.game_reports[p3.name]['disconnected'] = True
                        p3_action = -1
                else:
                    self.game_reports[p3.name]['disconnected'] = True
                    p3_action = -1

            else:
                self.run_func_w_time(
                    p1.restart, self.timeout, p1.name)
                p1_action = self.run_func_w_time(
                    p1.get_action, self.timeout, p1.name, -1)
                self.run_func_w_time(
                    p2.restart, self.timeout, p1.name)
                p2_action = self.run_func_w_time(
                    p2.get_action, self.timeout, p2.name, -1)
                self.run_func_w_time(
                    p3.restart, self.timeout, p1.name)
                p3_action = self.run_func_w_time(
                    p3.get_action, self.timeout, p3.name, -1)
            
            if p1_action not in self.valid_actions: 
                p1_action = -1
            if p2_action not in self.valid_actions: 
                p2_action = -1
            if p3_action not in self.valid_actions: 
                p3_action = -1

            self.game_reports[p1.name]['action_history'].append(p1_action)
            self.game_reports[p2.name]['action_history'].append(p2_action)
            self.game_reports[p3.name]['action_history'].append(p3_action)

            p1_util, p2_util, p3_util = self.calculate_utils(
                p1_action, p2_action, p3_action)
            self.game_reports[p1.name]['util_history'].append(p1_util)
            self.game_reports[p2.name]['util_history'].append(p2_util)
            self.game_reports[p3.name]['util_history'].append(p3_util)

            p1.game_report.game_history['my_action_history'].append(p1_action)
            p1.game_report.game_history['my_utils_history'].append(p1_util)
            p1.game_report.game_history['opp1_action_history'].append(p2_action)
            p1.game_report.game_history['opp1_utils_history'].append(p2_util)
            p1.game_report.game_history['opp2_action_history'].append(p3_action)
            p1.game_report.game_history['opp2_utils_history'].append(p3_util)

            p2.game_report.game_history['my_action_history'].append(p2_action)
            p2.game_report.game_history['my_utils_history'].append(p2_util)
            p2.game_report.game_history['opp1_action_history'].append(p1_action)
            p2.game_report.game_history['opp1_utils_history'].append(p1_util)
            p2.game_report.game_history['opp2_action_history'].append(p3_action)
            p2.game_report.game_history['opp2_utils_history'].append(p3_util)

            p3.game_report.game_history['my_action_history'].append(p3_action)
            p3.game_report.game_history['my_utils_history'].append(p3_util)
            p3.game_report.game_history['opp1_action_history'].append(p1_action)
            p3.game_report.game_history['opp1_utils_history'].append(p1_util)
            p3.game_report.game_history['opp2_action_history'].append(p2_action)
            p3.game_report.game_history['opp2_utils_history'].append(p2_util)

            if self.handin_mode:
                try:
                    self.run_func_w_time(p1.update, self.timeout, p1.name)
                except:
                    self.game_reports[p1.name]['disconnected'] = True

                try:
                    self.run_func_w_time(p2.update, self.timeout, p2.name)
                except:
                    self.game_reports[p2.name]['disconnected'] = True

                try:
                    self.run_func_w_time(p3.update, self.timeout, p3.name)
                except:
                    self.game_reports[p3.name]['disconnected'] = True
            else:
                self.run_func_w_time(p1.update, self.timeout, p1.name)
                self.run_func_w_time(p2.update, self.timeout, p2.name)
                self.run_func_w_time(p3.update, self.timeout, p3.name)

        if self.handin_mode:
            with open(self.logging_path, 'a') as file:
                file.write(f"Game {self.game_num}: {p1.name} VS {p2.name} VS {p3.name}\n")
        else: 
            print(f"Game {self.game_num}: {p1.name} VS {p2.name} VS {p3.name}")
            
        p1_tot_util = self.print_results(p1)
        p2_tot_util = self.print_results(p2)
        p3_tot_util = self.print_results(p3)
        agents = [p1, p2, p3]
        total_u = [p1_tot_util, p2_tot_util, p3_tot_util]
        winner = agents[np.argmax(total_u)].name
        self.results.append(
            [p1.name, p2.name, p3.name, p1_tot_util, p2_tot_util, p3_tot_util, winner]) 
        
        if self.handin_mode: 
            try:
                with open(self.shortcut_path, 'r', encoding='utf-8') as file:
                    data = json.load(file)
            except Exception as e:
                print(f"Error reading the file: {e}")
                data = {}
            
            names = [p1.name, p2.name, p3.name]
            timestamps = [p1.timestamp, p2.timestamp, p3.timestamp]
            combined = sorted(zip(names, timestamps, total_u), key=lambda x: x[0])
            sorted_names, sorted_timestamps, sorted_total_u = zip(*combined)
            data["|".join(sorted_names)] = {}
            data["|".join(sorted_names)]["util"] = [int(x) for x in sorted_total_u]
            data["|".join(sorted_names)]["timestamps"] = sorted_timestamps
            try:
                with open(self.shortcut_path, 'w', encoding='utf-8') as file:
                    json.dump(data, file, ensure_ascii=False, indent=4)
            except Exception as e:
                print(f"Error writing to the file: {e}")
                data = {}
                        
    def run_game(self, p1, p2, p3):
        if self.handin_mode:
            try:
                with open(self.shortcut_path, 'r') as file:
                    data = json.load(file)
            except:
                data = {}
            
            names = [p1.name, p2.name, p3.name]
            timestamps = [p1.timestamp, p2.timestamp, p3.timestamp]
            combined = sorted(zip(names, timestamps), key=lambda x: x[0])
            sorted_names, sorted_timestamps = zip(*combined)
            name_key = "|".join(sorted_names)
            
            if random.random() > 0.07 and name_key in data: 
                total_u = data[name_key]["util"]
                winner = sorted_names[np.argmax(total_u)]
                self.results.append(list(sorted_names) + total_u + [winner])
            else: 
                self.run_helper(p1, p2, p3)
        else: 
            self.run_helper(p1, p2, p3)
        
        self.game_num += 1
        self.reset_game_reports()

    def print_results(self, p):
        action_counts = [0 for _ in range(len(self.valid_actions) + 1)]
        for action in self.game_reports[p.name]['action_history']:
            if action in self.valid_actions:
                action_counts[action] += 1
            else:
                action_counts[len(self.valid_actions)] += 1
        print_smt = f"{p.name} set up their Lemonade Stand at"
        for i in range(len(self.valid_actions) - 1):
            print_smt += f" Location {i} {action_counts[i]} times,"
        print_smt += f" and Location {len(self.valid_actions) - 1} {action_counts[len(self.valid_actions) - 1]} times."
        if not self.handin_mode:
            print(print_smt)
            if action_counts[len(self.valid_actions)] > 0:
                print(
                    f"{p.name} submitted {action_counts[len(self.valid_actions)]} invalid moves")
            if self.game_reports[p.name]['global_timeout_count'] > 0:
                print(
                    f"{p.name} timed out {self.game_reports[p.name]['global_timeout_count']} times")
                self.game_reports[p.name]['global_timeout_count'] = 0

        total_util = sum(self.game_reports[p.name]['util_history'])

        avg_util = total_util / \
            len(self.game_reports[p.name]['util_history'])
        if not self.handin_mode:
            print(
                f"{p.name} got a total utility of {total_util} and a average utility of {avg_util}")
        return total_util

    def summarize_results(self):
        import pandas as pd
        df = pd.DataFrame(self.results)
        df.columns = ['Agent 1', 'Agent 2', 'Agent 3',
                      'Agent 1 Score', 'Agent 2 Score', 'Agent 3 Score', 'Winner']

        if self.handin_mode: 
            with open(self.logging_path, 'a') as file:
                file.write(f"Extended Results: \n {df}")
        else: 
            print(f"Extended Results: \n {df}")

        total_util_dict = defaultdict(lambda: [0, 0])
        for p1, p2, p3, p1_util, p2_util, p3_util, _ in self.results:
            total_util_dict[p1][0] += p1_util
            total_util_dict[p2][0] += p2_util
            total_util_dict[p3][0] += p3_util
            total_util_dict[p1][1] += 1
            total_util_dict[p2][1] += 1
            total_util_dict[p3][1] += 1

        res_summary = []
        for key, value in total_util_dict.items():
            if self.handin_mode and self.game_reports[key]['disconnected']:
                res_summary.append([key, float('-inf'), float('-inf')])
            elif value[1] > 0:
                res_summary.append(
                    [key, value[0] / value[1], value[0] / (value[1] * self.num_rounds)])
            else:
                res_summary.append([key, 0, 0])

        sum_df = pd.DataFrame(res_summary)
        sum_df.columns = ['Agent Name', 'Average Utility', 'Final Score']
        sum_df = sum_df.sort_values('Final Score', ascending=False)
        sum_df = sum_df[sum_df['Final Score'] != float('-inf')]
        
        if not self.handin_mode:
            print(f"Results: \n {sum_df}")
        else: 
            today_date_formatted = datetime.now().strftime('%Y-%m-%d_%H%M')
            final_str = f"{today_date_formatted}\t"
            result_list = [str(item) for pair in zip(sum_df['Agent Name'], sum_df['Final Score']) for item in pair]
            final_str += "\t".join(result_list)
            print(final_str)
            with open(pkg_resources.resource_filename('agt_server', self.save_path), 'a') as file:
                file.write(final_str + "\n")
            
        return sum_df



from agt_server.agents.base_agents.lemonade_agent import LemonadeAgent
from agt_server.local_games.lemonade_arena import LemonadeArena
import argparse
import random
import numpy as np


class E2Agent(LemonadeAgent):
    def setup(self):
        self.gamma = 0.75
        self.rho = 0.5
        self.tol = 0.1
        self.stick = 5
        self.t = 0
        self.carrot_stick = 0
        self.prev_action = random.randint(0, 11)

    def distance(self, loc1, loc2):
        return min((loc1 - loc2) % 12, (loc2 - loc1) % 12) ** self.rho

    def opposite(self, loc):
        return (loc + 6) % 12

    def teehee(self, idx, time_idx):
        if (idx == 0):
            return self.get_action_history()[time_idx]
        elif (idx == 1):
            return self.get_opp1_action_history()[time_idx]
        else:
            return self.get_opp2_action_history()[time_idx]

    def compute_indices(self):
        """
        Trust wont be called before self.t is invalid heh
        """
        # stick indices
        self.gammas = []
        cur_gamma = 1
        gamma_sum = 0
        for i in range(1, self.t):
            self.gammas.append(cur_gamma)
            gamma_sum += cur_gamma
            cur_gamma *= self.gamma
        self.gammas.reverse()

        s1, s2 = (0, 0)
        for i in range(1, self.t):
            s1 -= self.gammas[i - 1] / gamma_sum * self.distance(self.get_opp1_action_history()[i],
                                                                 self.get_opp1_action_history()[i - 1])
            s2 -= self.gammas[i - 1] / gamma_sum * self.distance(self.get_opp2_action_history()[i],
                                                                 self.get_opp2_action_history()[i - 1])

        # now compute
        follow_cross = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        for i in range(3):
            for j in range(3):
                if (i == j):
                    continue
                for k in range(1, self.t):
                    i_act = self.teehee(i, k)
                    j_act = self.teehee(j, k - 1)
                    follow_cross[i][j] -= self.gammas[k - 1] / \
                        gamma_sum * self.distance(i_act, self.opposite(j_act))
        f1, f2 = (0, 0)

        for i in range(1, self.t):
            f1 -= self.gammas[i - 1] / gamma_sum * min(
                self.distance(self.teehee(1, i), self.opposite(
                    self.teehee(0, i - 1))),
                self.distance(self.teehee(1, i),
                              self.opposite(self.teehee(2, i - 1)))
            )
            f2 -= self.gammas[i - 1] / gamma_sum * min(
                self.distance(self.teehee(2, i), self.opposite(
                    self.teehee(0, i - 1))),
                self.distance(self.teehee(2, i),
                              self.opposite(self.teehee(1, i - 1)))
            )
        return s1, s2, follow_cross, f1, f2

    def calc_bias(self, target_opp, stick_opp_loc):
        acts = None
        if target_opp == 1:
            acts = self.get_opp1_action_history()
        else:
            acts = self.get_opp1_action_history()

        # calculate a bias
        opp = self.opposite(stick_opp_loc)
        bias = 0
        for i in range(len(acts)):
            if i > opp or i < acts[i]:
                # its on the "left"
                bias -= 1
            elif i < opp or i > acts[i]:
                bias += 1
        return bias

    def get_action(self):
        if self.t > 0:
            last_opp_1 = self.get_opp1_action_history()[-1]
            last_opp_2 = self.get_opp2_action_history()[-1]
            prev_util = self.get_util_history()[-1]

        if self.stick > 0:
            self.stick -= 1
            return self.prev_action

        if self.carrot_stick > 0:
            # move yourself to the side a bit
            self.carrot_stick -= 1
            if self.carrot_stick > 3:
                return self.prev_action
            if self.bias:
                self.prev_action = (self.prev_action + 1) % 12
            else:
                self.prev_action = (self.prev_action - 1) % 12
            return self.prev_action

        s1, s2, follow_cross, f1, f2 = self.compute_indices()
        if s1 > max(max(f1, s2), f2) + self.tol:
            self.prev_action = self.opposite(last_opp_1)
        elif s2 > max(max(f1, s1), f2) + self.tol:
            self.prev_action = self.opposite(last_opp_2)
        elif s1 > f1 + self.tol and s2 > f2 + self.tol:
            if (prev_util > 8):
                self.stick += 1
            elif (s1 > s2):
                self.prev_action = self.opposite(last_opp_1)
            else:
                self.prev_action = self.opposite(last_opp_2)
        elif f1 > max(max(s1, s2), f2) + self.tol:
            if (follow_cross[1][0] < follow_cross[1][2]):
                # 1 is following 2! sit on 2
                self.prev_action = last_opp_2
        elif f2 > max(max(s1, s2), f1) + self.tol:
            if (follow_cross[2][0] < follow_cross[2][1]):
                # 2 is following 1! sit on 1
                self.prev_action = last_opp_1
        elif f1 > s1 + self.tol and f2 > s2 + self.tol and follow_cross[1][2] > follow_cross[1][0] and follow_cross[2][
                1] > follow_cross[2][0]:
            # probably following each other
            if f1 > f2:
                self.prev_action = last_opp_1
            else:
                self.prev_action = last_opp_2
        elif last_opp_1 == self.opposite(last_opp_2):
            # they are opposite!
            # carrot and stick
            self.carrot_stick = 8
            if (f1 > f2):
                # f1 is less sticky - so just sit on top of it and force to move :D
                # but want to force it to collaborate with me... idk im too lazy to impl this part
                self.prev_action = last_opp_1
                self.bias = self.calc_bias(1, last_opp_2)
            else:
                self.prev_action = last_opp_2
                self.bias = self.calc_bias(2, last_opp_1)

        return self.prev_action

    def update(self):
        self.t += 1

################### SUBMISSION #####################
agent_submission = E2Agent("E2A")
####################################################


if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = E2Agent(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = LemonadeArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                E2Agent("Agent_1"),
                E2Agent("Agent_2"),
                E2Agent("Agent_3"),
                E2Agent("Agent_4")
            ]
        )
        arena.run()



from agt_server.server.games.game import Game
import logging
import json
import traceback
import asyncio

class SuppressSocketSendError(logging.Filter):
    def filter(self, record):
        return "socket.send() raised exception" not in record.getMessage()

logger = logging.getLogger('asyncio')
logger.addFilter(SuppressSocketSendError())
class LemonadeGame(Game):
    def __init__(self, num_rounds=1000, player_data=[], player_types=[], permissions_map={}, game_kick_timeout=60, game_name=None, invalid_move_penalty=0, timeout_tolerance=10):
        super().__init__(num_rounds, player_data,
                         player_types, permissions_map, game_kick_timeout, game_name, invalid_move_penalty, timeout_tolerance)
        for data in self.player_data:
            self.game_reports[data['address']] = {
                "action_history": [],
                "util_history": [],
                "timeout_count": 0,
                "global_timeout_count": 0,
                "disconnected": False,
                "disqualification_message": ""
            }
        self.valid_actions = list(range(12))

    def calculate_utils(self, p1_action, p2_action, p3_action):
        utils = [0, 0, 0]
        if p1_action not in self.valid_actions and p2_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [0, 0, 0]
        elif p1_action not in self.valid_actions and p2_action not in self.valid_actions:
            utils = [self.invalid_move_penalty, self.invalid_move_penalty, 24]
        elif p1_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [self.invalid_move_penalty, 24, self.invalid_move_penalty]
        elif p2_action not in self.valid_actions and p3_action not in self.valid_actions:
            utils = [24, self.invalid_move_penalty, self.invalid_move_penalty]
        elif p1_action not in self.valid_actions:
            utils = [self.invalid_move_penalty, 12, 12]
        elif p2_action not in self.valid_actions:
            utils = [12, self.invalid_move_penalty, 12]
        elif p3_action not in self.valid_actions:
            utils = [12, 12, self.invalid_move_penalty]
        elif p1_action == p2_action and p2_action == p3_action:
            utils = [8, 8, 8]
        elif p1_action == p2_action:
            utils = [6, 6, 12]
        elif p1_action == p3_action:
            utils = [6, 12, 6]
        elif p2_action == p3_action:
            utils = [12, 6, 6]
        else:
            actions = [p1_action, p2_action, p3_action]
            sorted_actions = sorted(actions)
            index_map = {action: index for index,
                         action in enumerate(actions)}
            sorted_indices = [index_map[action]
                              for action in sorted_actions]
            u1 = sorted_actions[1] - sorted_actions[0]
            u2 = sorted_actions[2] - sorted_actions[1]
            u3 = 12 + sorted_actions[0] - sorted_actions[2]
            utils[sorted_indices[0]] = u1 + u3
            utils[sorted_indices[1]] = u1 + u2
            utils[sorted_indices[2]] = u2 + u3
        return utils

    def simulate_round(self):
        p1 = self.player_data[0]['address']
        p2 = self.player_data[1]['address']
        p3 = self.player_data[2]['address']
        if self.game_reports[p1]['disconnected'] or self.game_reports[p2]['disconnected'] or self.game_reports[p3]['disconnected']:
            p1_util, p2_util, p3_util = 0, 0, 0
        else:
            p1_action = self.game_reports[p1]['action_history'][-1]
            p2_action = self.game_reports[p2]['action_history'][-1]
            p3_action = self.game_reports[p3]['action_history'][-1]
            p1_util, p2_util, p3_util = self.calculate_utils(
                p1_action, p2_action, p3_action)
        self.game_reports[p1]['util_history'].append(p1_util)
        self.game_reports[p2]['util_history'].append(p2_util)
        self.game_reports[p3]['util_history'].append(p3_util)

    async def run_game(self):
        for round in range(self.num_rounds):
            for i in range(len(self.player_data)):
                data = self.player_data[i]
                player_type = self.player_types[i]
                writer, reader = data['client']
                message = {"message": "send_preround_data",
                           "player_type": player_type}

                if not self.game_reports[data['address']]['disconnected'] == True:
                    try:
                        # # LOGGING: Delete this
                        # print(f"Asking if {data['name']} is ready for preround data", flush=True)
                        writer.write(json.dumps(message).encode())
                        await writer.drain()
                        resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                        
                        resp = json.loads(resp)
                        assert resp['message'] == 'preround_data_recieved', f"{data['name']} did not recieve preround_data"
                        # # LOGGING: Delete this
                        # print(f"{data['name']} recieved preround data", flush=True)
                    except asyncio.TimeoutError:
                        self.game_reports[data['address']
                                          ]['disqualification_message'] = f"{data['name']} Disqualified: Agent has timed out past server limits"
                        self.game_reports[data['address']
                                          ]['disconnected'] = True
                    except Exception as e:
                        stack_trace = traceback.format_exc()
                        self.game_reports[data['address']
                                          ]['disqualification_message'] = f"{data['name']} Disqualified: Agent failed to confirm readyness for the round. {type(e).__name__}: {e} \nStack Trace:\n{stack_trace}"
                        self.game_reports[data['address']
                                          ]['disconnected'] = True

            for data in self.player_data:
                if self.game_reports[data['address']]['timeout_count'] < self.timeout_tolerance:
                    writer, reader = data['client']
                    message = {"message": "request_action"}
                    if not self.game_reports[data['address']]['disconnected'] == True:
                        try:
                            # # LOGGING: Delete this
                            # print(f"Asking {data['name']} for an action", flush=True)
                            writer.write(json.dumps(message).encode())
                            await writer.drain()
                            resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                            
                            if not resp:
                                # # LOGGING: Delete this
                                # print(f"{data['name']} gave no response", flush=True)
                                self.game_reports[data['address']
                                          ]['disqualification_message'] = f"{data['name']} Disqualified: Agent sent a blank action response"
                                self.game_reports[data['address']
                                                  ]['disconnected'] = True
                                self.game_reports[data['address']
                                                  ]['action_history'].append(-1)
                            else:
                                resp = json.loads(resp)
                                if resp and resp['message'] == 'provide_action':
                                    if resp['timeout']: 
                                        print(f"{data['name']} has timed out")
                                        self.game_reports[data['address']
                                              ]['action_history'].append(-1)
                                        self.game_reports[data['address']
                                                        ]['timeout_count'] += 1
                                        self.game_reports[data['address']
                                                        ]['global_timeout_count'] += 1
                                    else: 
                                        # # LOGGING: Delete this
                                        # print(f"{data['name']} sucessfully gave action {resp['action']}", flush=True)  
                                        self.game_reports[data['address']
                                                      ]['action_history'].append(resp['action'])
                        except asyncio.TimeoutError:
                            # # LOGGING: Delete this
                            # print(f"{data['name']} has timed out", flush=True)  
                            self.game_reports[data['address']
                                          ]['disqualification_message'] = f"{data['name']} Disqualified: Agent has timed out past server limits"
                            self.game_reports[data['address']
                                              ]['disconnected'] = True
                            self.game_reports[data['address']
                                              ]['action_history'].append(-1)
                        except Exception as e:
                            # # LOGGING: Delete this
                            # print(f"{data['name']} has other error", flush=True)  
                            stack_trace = traceback.format_exc()
                            self.game_reports[data['address']
                                          ]['disqualification_message'] = f"{data['name']} Disqualified: Agent had a socket error. {type(e).__name__}: {e} \nStack Trace:\n{stack_trace}"
                            self.game_reports[data['address']
                                              ]['disconnected'] = True
                            self.game_reports[data['address']
                                              ]['action_history'].append(-1)
                    else:
                        # # LOGGING: Delete this
                        # print(f"{data['name']} has timed out too much", flush=True)  
                        self.game_reports[data['address']
                                          ]['disqualification_message'] = f"{data['name']} Disqualified: Agent timed out {self.timeout_tolerance} times in a row"
                        self.game_reports[data['address']
                                          ]['disconnected'] = True
                        self.game_reports[data['address']
                                          ]['action_history'].append(-1)
            self.simulate_round()

            if round != self.num_rounds - 1:
                for i in range(len(self.player_data)):
                    data = self.player_data[i]
                    player_type = self.player_types[i]
                    opp1_data = self.player_data[(1 - i) % 3]
                    opp2_data = self.player_data[(2 - i) % 3]
                    writer, reader = data['client']

                    # EDIT THIS TO ADD PERMISSIONS
                    if 'all' in self.permissions_map[player_type] and self.permissions_map[player_type]['all']:
                        message = {"message": "prepare_next_round",
                                   "player_type": player_type,
                                   "permissions": ['all'],
                                   "opp_names": [opp1_data['name'], opp2_data['name']],
                                   "my_action": self.game_reports[data['address']]['action_history'][-1],
                                   "my_utils": self.game_reports[data['address']]['util_history'][-1],
                                   "opp1_action":  self.game_reports[opp1_data['address']]['action_history'][-1],
                                   "opp1_utils": self.game_reports[opp1_data['address']]['util_history'][-1],
                                   "opp2_action":  self.game_reports[opp2_data['address']]['action_history'][-1],
                                   "opp2_utils": self.game_reports[opp2_data['address']]['util_history'][-1],
                                   }
                    else:
                        message = {"message": "prepare_next_round",
                                   "player_type": player_type,
                                   "permissions": [],
                                   }
                        if 'my_action' in self.permissions_map[player_type] and self.permissions_map[player_type]['my_action']:
                            message['my_action'] = self.game_reports[data['address']
                                                                     ]['action_history'][-1]
                            message['permissions'].append('my_action')
                        if 'my_utils' in self.permissions_map[player_type] and self.permissions_map[player_type]['my_utils']:
                            message['my_utils'] = self.game_reports[data['address']
                                                                    ]['util_history'][-1]
                            message['permissions'].append('my_utils')
                        if 'opp1_action' in self.permissions_map[player_type] and self.permissions_map[player_type]['opp1_action']:
                            message['opp1_action'] = self.game_reports[opp1_data['address']
                                                                       ]['action_history'][-1]
                            message['permissions'].append('opp1_action')
                        if 'opp1_utils' in self.permissions_map[player_type] and self.permissions_map[player_type]['opp1_utils']:
                            message['opp1_utils'] = self.game_reports[opp1_data['address']
                                                                      ]['util_history'][-1]
                            message['permissions'].append('opp1_utils')

                        if 'opp2_action' in self.permissions_map[player_type] and self.permissions_map[player_type]['opp2_action']:
                            message['opp2_action'] = self.game_reports[opp2_data['address']
                                                                       ]['action_history'][-1]
                            message['permissions'].append('opp2_action')
                        if 'opp2_utils' in self.permissions_map[player_type] and self.permissions_map[player_type]['opp2_utils']:
                            message['opp2_utils'] = self.game_reports[opp2_data['address']
                                                                      ]['util_history'][-1]
                            message['permissions'].append('opp2_utils')

                    message['global_timeout_count'] = self.game_reports[data['address']
                                                                        ]['global_timeout_count']
                    
                    if not self.game_reports[data['address']]['disconnected'] == True:
                        try:
                            # # LOGGING: Delete this
                            # print(f"Asking if {data['name']} is ready", flush=True)
                            writer.write(json.dumps(message).encode())
                            await writer.drain()
                            resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                            
                            resp = json.loads(resp)
                            assert resp['message'] == 'ready_next_round', f"{data['name']} was not ready for the next round"
                            # # LOGGING: Delete this
                            # print(f"{data['name']} is confirmed ready", flush=True)
                        except asyncio.TimeoutError:
                            self.game_reports[data['address']
                                          ]['disqualification_message'] = f"{data['name']} Disqualified: Agent has timed out past server limits"
                            self.game_reports[data['address']
                                              ]['disconnected'] = True
                        except Exception as e:
                            stack_trace = traceback.format_exc()
                            self.game_reports[data['address']
                                          ]['disqualification_message'] = f"{data['name']} Disqualified: Agent was not ready for the next_round. {type(e).__name__}: {e} \nStack Trace:\n{stack_trace}"
                            self.game_reports[data['address']
                                              ]['disconnected'] = True
                    else:
                        message = {"message": "disqualified",
                                   "disqualification_message": self.game_reports[data['address']]['disqualification_message']}
                        try:
                            # # LOGGING: Delete this
                            # print(f"Telling {data['name']} is disqualified", flush=True)  
                            writer.write(json.dumps(message).encode())
                            await writer.drain()
                        except Exception as e:
                            # # LOGGING: Delete this
                            # print(f"{data['name']} failed to recieve message, {type(e).__name__}: {e}", flush=True)  
                            continue

        message = {"message": "prepare_next_game"}
        for data in self.player_data:
            if not self.game_reports[data['address']]['disconnected'] == True:
                try:
                    writer, reader = data['client']
                    # # LOGGING: Delete this
                    # print(f"Telling {data['name']} to prepare next game", flush=True)  
                    writer.write(json.dumps(message).encode())
                    await writer.drain()
                    resp = await asyncio.wait_for(reader.read(1024), timeout=self.kick_time)
                    
                    resp = json.loads(resp)
                    assert resp['message'] == 'ready_next_game', f"{data['name']} was not prepared for next game"
                    # # LOGGING: Delete this
                    # print(f"{data['name']} was prepared for the next game", flush=True)  
                except asyncio.TimeoutError:
                    self.game_reports[data['address']
                                    ]['disqualification_message'] = f"{data['name']} Disqualified: Agent has timed out past server limits"
                    self.game_reports[data['address']
                                        ]['disconnected'] = True
                except Exception as e:
                    stack_trace = traceback.format_exc()
                    self.game_reports[data['address']
                                    ]['disqualification_message'] = f"{data['name']} Disqualified: Agent was not ready for the next game. {type(e).__name__}: {e} \nStack Trace:\n{stack_trace}"
                    self.game_reports[data['address']
                                        ]['disconnected'] = True

        return self.game_reports



import argparse
import numpy as np
from agt_server.agents.base_agents.rps_agent import RPSAgent
from agt_server.local_games.rps_arena import RPSArena
from agt_server.agents.test_agents.rps.ta_agent.my_agent import TAAgent


class FictitiousPlayAgent(RPSAgent):
    def setup(self):
        self.ROCK, self.PAPER, self.SCISSORS = 0, 1, 2
        self.actions = [self.ROCK, self.PAPER, self.SCISSORS]
        self.opp_action_history = []

    def get_action(self):
        dist = self.predict()
        best_move = self.optimize(dist)
        return self.actions[best_move]

    def update(self):
        """
        Updates opp action history to be a record of opponent moves
        Rock - 0, Scissors - 1, Paper - 2
        """
        self.opp_action_history = self.get_opp_action_history()

    def predict(self):
        """
        Uses the opponent’s previous moves (self.opp_action_history) to generate and save a probability distribution
        over the opponent’s next move in dist.
        """
        print(self.opp_action_history)
        dist = np.zeros(len(self.actions))
        for a in self.opp_action_history:
            dist[a] += 1
        if sum(dist) == 0:
            return np.ones(len(self.actions))/len(self.actions)
        return dist/sum(dist)

    def optimize(self, dist):
        """
        Given the distribution over the opponent's next move (output of predict) and knowledge of the payoffs (self.calculate_utils),
        Return the best move according to Ficticious Play.
        Please return one of [self.ROCK, self.PAPER, self.SCISSORS]
        """
        # TODO Calculate the expected payoff of each action and return the action with the highest payoff
        action_utils = np.zeros(len(self.actions))
        for i, a1 in enumerate(self.actions):
            # Calculate the payoff
            for j, a2 in enumerate(self.actions):
                action_utils[i] += dist[j]* self.calculate_utils(a1, a2)[0]

        best_action = np.argmax(action_utils)
        return best_action

agent_submission = FictitiousPlayAgent("FP")

if __name__ == "__main__":
    #### DO NOT TOUCH THIS #####
    parser = argparse.ArgumentParser(description='My Agent')
    parser.add_argument('agent_name', type=str, help='Name of the agent')
    parser.add_argument('--join_server', action='store_true',
                        help='Connects the agent to the server')
    parser.add_argument('--ip', type=str, default='127.0.0.1',
                        help='IP address (default: 127.0.0.1)')
    parser.add_argument('--port', type=int, default=8080,
                        help='Port number (default: 8080)')

    args = parser.parse_args()

    agent = FictitiousPlayAgent(args.agent_name)
    if args.join_server:
        agent.connect(ip=args.ip, port=args.port)
    else:
        arena = RPSArena(
            num_rounds=1000,
            timeout=1,
            players=[
                agent,
                TAAgent("TA_Agent_1"),
                TAAgent("TA_Agent_2"),
                TAAgent("TA_Agent_3"),
                TAAgent("TA_Agent_4")
            ]
        )
        arena.run()
